(**************************************************************************)
(*                                                                        *)
(*  Copyright (C) 2013                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* Calmos plugin for analysis of the UART of the Leon *)
open Property_status.Feedback
open Cil
open Cil_types

(***** Registering the plugin *****)

module M = Plugin.Register(
  struct
    let name = "calmos script"
    let shortname = "calmos"
    let help = "Calmos Leon analysis script"
  end)

(***** The various command-line options *****)

module Restrict = M.True(
  struct
    let option_name = "-restrict"
    let help =
      "when on (default), uses only a restricted set of driver commands"
  end)

module MallocType = M.String(
  struct
    let option_name = "-malloc-type"
    let default = "INDIVIDUAL"

    let help = "malloc implementation to use during analysis \
                <s> can be INDIVIDUAL, INFINITE, CHUNKS, HEAP or POSITION, \
                defaults to " ^default
    let arg_name = "s"
  end)

module Force = M.False(
  struct
    let option_name = "-force"
    let help = "force constant propagation analysis even if generated file \
                 is up-to-date (default: no)"
  end
)

module FramaCPathFinder = M.False(
  struct
    let option_name = "-frama-c-path-finder"
    let help = "use Frama-C to find a path instead of some external tool \
                 (default: no)"
  end)

module UseLtl = M.False(
  struct
    let option_name = "-use-ltl"
    let help = "use an ltl formula as scenario description (default: use a \
                 ya automaton)"
  end
)

module OnlyFinalAnalysis = M.False(
  struct
    let option_name = "-only-final"
    let help = "only run Frama-C on scenario.c"
end)

module Max_nb_scenarii = M.Int(
  struct
    let option_name = "-max-nb-scenarii"
    let default = 1
    let arg_name = "n"
    let help = "output <n> distinct scenarii \
                leading to the alarm (default is 1)"
  end)

(** Use non free builtins functions in value *)
module Use_builtins = M.False(
  struct
    let option_name = "-use-builtins"
    let help = "use value analysis builtins for memory operations (malloc &al)"
  end)

let () = Max_nb_scenarii.set_range ~min:1 ~max:max_int

let dkey_time = M.register_category "time-stats"

(* New projects gets most of their state reset to the default value. This is
   in particular the case for options. We need thus to pass our options that
   are used in later stages from a project to another.
 *)
let keep_calmos_options src dest =
  let selection =
    State_selection.of_list
      [UseLtl.self; FramaCPathFinder.self; Max_nb_scenarii.self;
       Use_builtins.self ]
  in
  Project.copy ~selection ~src dest;
  (* Similarly, certain command-lines options for other plugins need
     to be carried on across project changes. *)
  let opt = Project.on src Dynamic.Parameter.Bool.get "-aorai-dot-sep-labels" ()
  in Project.on dest (Dynamic.Parameter.Bool.set "-aorai-dot-sep-labels") opt

(***** Source files *****)

(* the directory where to find Frama-C's libc. *)
let libc_path = Filename.concat Config.datadir "libc"

(* source files *)
let src_files () =
  let files =
    ["leon-no-asm.c"; "serial_core.c"; "tty_io.c"; "stubs.c"; "n_tty.c"]
  in
  let libc_files =
    [  "string.c";  "stdlib.c"; "__fc_builtin.c"]
  in
  if Use_builtins.get () then
    let libc_files = List.map (Filename.concat libc_path) libc_files in
    files @ libc_files
  else
    (Filename.concat Config.datadir "libc.c") :: files

(***** Configuration *****)
(* functions in which value analysis must not keep states separated. *)
let nounroll = ["tty_buffer_alloc"; "tty_insert_flip_string";
                "tty_buffer_request_room"; "tty_buffer_find"]

(* configuration of the value analysis plugin *)
let value_config () =
  Dynamic.Parameter.Int.set "-slevel" 10000;
  Dynamic.Parameter.Int.set "-value-verbose" 0;
  if Use_builtins.get () then begin
    List.iter (fun s->
      Dynamic.Parameter.StringSet.set "-slevel-function"
        (Datatype.String.Set.add (s^":0")
           (Dynamic.Parameter.StringSet.get "-slevel-function" ())))
      nounroll
  end

(* general configuration for parsing files *)
let config () =
  let drestrict = if Restrict.get () then " -DRESTRICT" else "" in
  let dbuiltin = if Use_builtins.get() then "" else " -DNO_BUILTIN" in
  Kernel.CppCommand.set
    ("gcc -CC -E -I" ^ libc_path ^
       " -I. -D__KERNEL__ -DCONFIG_LEON -DCONFIG_SERIAL_CORE_CONSOLE \
       -DFRAMA_C_MALLOC_" ^ (MallocType.get()) ^
       drestrict ^ dbuiltin);
  Kernel.PreprocessAnnot.set true;
  Kernel.SimplifyCfg.set true;
  Kernel.Files.set (src_files ());
  value_config()
;;

(***** AST manipulations *****)

exception Found of enumitem

(* retrieve the enum items corresponding to the states of the
   automaton. *)
let find_state name =
  let ast = Ast.get () in
  try
    List.iter
      (function
           GEnumTag(enuminfo,_) when enuminfo.ename = "aorai_States" ->
             let x = List.find (fun x -> x.einame = name) enuminfo.eitems in
             raise (Found x)
         | GEnumTag(enuminfo,_) -> M.debug "found enum %s" enuminfo.ename
         | _ -> ())
      ast.globals;
    M.fatal "unable to retrieve enum aorai_States"
  with Found x -> x
    | Not_found ->
        M.fatal "generated automaton does not have a state named %s" name

(* pretty-printing of scenario *)

let operation_of_integer i =
  let i = Integer.to_int i in
  match i with
    | 0 -> "tiocmset"
    | 1 -> "write"
    | 2 -> "put_char"
    | 3 -> "flush_chars"
    | 4 -> "write_room"
    | 5 -> "chars_in_buffer"
    | 6 -> "throttle"
    | 7 -> "unthrottle"
    | 8 -> "flush_buffer"
    | 9 -> "send_xchar"
    | 10 -> "wait_until_sent"
    | 11 -> "start"
    | 12 -> "stop"
    | 13 -> "hangup"
    | 14 -> "hw interruption"
    | _ -> "unknown operation (" ^ string_of_int i ^ ")"

let uart_status_of_integer i =
  let i = Integer.to_int i in
  let add r s = if r = "" then s else r ^ "|" ^ s in
  let check_flag f = i land f <> 0 in
  let res = "" in
  let res = if check_flag 0x01 then add res "DR" else res in
  let res = if check_flag 0x02 then add res "TS" else res in
  let res = if check_flag 0x04 then add res "TH" else res in
  let res = if check_flag 0x08 then add res "BR" else res in
  let res = if check_flag 0x10 then add res "OV" else res in
  let res = if check_flag 0x20 then add res "PE" else res in
  let res = if check_flag 0x40 then add res "FE" else res in
  let res =
    if check_flag (0x10 lor 0x20 lor 0x40) then
      add res "ERROR"
    else res
  in
  res

(* Check if the current state of value analysis has reached the alarm
   state. This is the hook that will be called during the last phase
   of the analysis.
 *)
let inspect_automaton_state () =
  (* retrieve the variables whose value is important *)
  (* state of the automaton *)
  let nb_scenario_found = ref 0 in
  let final_state = if UseLtl.get () then "accept_S4" else "aorai_CurStates" in
  let state_var =
    Globals.Vars.find_from_astinfo final_state VGlobal
  in
  let state_expr = Var state_var, NoOffset in
  let comparison =
    if UseLtl.get () then
      Integer.one
    else
      let ei = find_state "alarm" in
      match (Cil.constFold true ei.eival).enode with
        | Const (CInt64 (v,_,_)) -> v
        | _ -> M.fatal "alarm enumitem has not a compile-time constant value"
  in
  let comparison = Ival.inject_singleton comparison in
  (* sequence of calls. *)
  let count = Globals.Vars.find_from_astinfo "calmos_count" VGlobal in
  let scenario = Globals.Vars.find_from_astinfo "calmos_scenario" VGlobal in
  (* sequence of hardware status. *)
  let status_scenario = Globals.Vars.find_from_astinfo "status_scenario"
    VGlobal
  in
  let status_count = Globals.Vars.find_from_astinfo "status_count" VGlobal
  in
  let with_alarms=CilE.warn_none_mode in
  (* the callback that will be registered for the value analysis. *)
  fun (call_stack, states) ->
    M.feedback ~level:2 "entering Calmos hook";
    let (kf,_ki) =
      try List.hd call_stack with Invalid_argument _ ->
        M.fatal "outside of program execution!"
    in
    (* The check itself is performed for the special function calmos_hook,
       called at the end of each step of the main loop. *)
    if Kernel_function.get_name kf = "calmos_hook" then begin
      M.feedback ~level:2 "Inspecting automaton state";
      (* We see if we have reached the alarm state. *)
      let eval_one_state state =
        let loc = Cil_datatype.Location.unknown in
        M.feedback ~level:2 "state is@\n%a" Cvalue.Model.pretty state;
        let _,a_state = !Db.Value.eval_lval ~with_alarms None state state_expr
        in
        M.feedback ~level:2 "alarm state contains: %a"
          Cvalue.V.pretty a_state;
        let _,int_val = Cvalue.V.find_lonely_key a_state in
        if Ival.equal comparison int_val then begin
          M.feedback "Reached alarm state";
          (* compute the sequence of calls * status *)
          let rec make_scenario i l =
            let _, call_nb = !Db.Value.eval_lval ~with_alarms None
              state (Var scenario, Index(integer ~loc (i+1),NoOffset))
            in
            let _, status = !Db.Value.eval_lval ~with_alarms None
              state (Var status_scenario, Index (integer ~loc i, NoOffset))
            in
            M.feedback ~level:2 "step %d contains %a, UART status is %a" i
              Cvalue.V.pretty call_nb Cvalue.V.pretty status;
            let _, int_val = Cvalue.V.find_lonely_key call_nb in
            let int_val = Ival.project_int int_val in
            let _, status = Cvalue.V.find_lonely_key status in
            let status = Ival.project_int status in
            if i = 0 then (int_val, status) :: l
            else make_scenario (i-1) ((int_val, status) :: l)
          in
          (* Check number of steps of the scenario. *)
          let _,steps = !Db.Value.eval_lval ~with_alarms None
            state (Var count,NoOffset)
          in
          let _,status_steps =  !Db.Value.eval_lval ~with_alarms None
            state (Var status_count,NoOffset)
          in
          M.feedback ~level:2 "number of steps in scenario: %a"
            Cvalue.V.pretty steps;
          let _,steps =  Cvalue.V.find_lonely_key steps in
          (* The first call of calmos_hook only registers the
             initial state of the automaton
           *)
          let steps = Integer.pred (Ival.project_int steps) in
          let _, status_steps = Cvalue.V.find_lonely_key status_steps in
          let status_steps = Ival.project_int status_steps in
          if not (Integer.equal status_steps steps) then
            M.warning
              "found different values for calmos_count and status_count";
          let steps = Integer.min status_steps steps in

          let scenario = make_scenario (Integer.to_int steps -1) [] in
          (* pretty-printing of the scenario *)
          let pretty_print_step fmt (i, j) =
            Format.fprintf fmt "call operation %s, UART status is %s"
              (operation_of_integer i)
              (uart_status_of_integer j)
          in
          M.feedback "Sequence of calls:@\n%a"
            (Pretty_utils.pp_list ~sep:"@\n" pretty_print_step)
            scenario;
          (* If we have reached the max number of scenarii asked for by the
             user, we can now exit. *)
          incr nb_scenario_found;
          if !nb_scenario_found >= Max_nb_scenarii.get() then
            raise Cmdline.Exit
        end
      in
      (try
         let ki = Kernel_function.find_return kf in
         let states = Lazy.force states in
         let states = Cil_datatype.Stmt.Hashtbl.find states ki in
         List.iter eval_one_state states
       with
         | Not_found ->
             (* This exception is raised if the return statement
                of calmos_hook has not been reached. This is the
                case when the assertion in the function is
                invalid, which means that automaton state is
                unchanged from previous call. No need to inspect
                state again.  *)
             M.feedback ~level:2 "In a dead-end"

         | Ival.Not_Singleton_Int ->
             (* This exception is raised if the value analysis
                starts to become imprecise for the automaton
                state. In this case, we won't be able to
                conclude. *)
             M.warning
               "Analysis is not precise enough to conclude."
         | Cmdline.Exit as exn -> raise exn
             (* this is normal. we have finished the analysis and can exit
                normally *)
         | exn ->
             M.error "Unexpected exception: %s" (Printexc.to_string exn);
             Printexc.print_backtrace stdout;
             raise exn);
    end else begin
      (* We are not at the end of a step of the main loop, we can
         at least check that the value analysis still has precise
         values for the states of the automaton. Otherwise, we can
         directly abort the current analysis, as it won't be able
         to give anything useful. *)
      M.feedback ~level:2 "Not the good function";
      (try
         let ki = Kernel_function.find_first_stmt kf in
         let states = Lazy.force states in
         let state_init =
           try Cil_datatype.Stmt.Hashtbl.find states ki
           with Not_found ->
             M.warning "No state associated to entry point of function %s: %a"
               (Kernel_function.get_name kf) Printer.pp_stmt ki;
             raise Exit
         in
         let state_end =
           let ki = Kernel_function.find_return kf in
           try
             Cil_datatype.Stmt.Hashtbl.find states ki
           with Not_found ->
             M.warning "No state associated to return stmt of function %s: %a"
               (Kernel_function.get_name kf) Printer.pp_stmt ki;
             raise Exit
         in
         try
           let states =
             if UseLtl.get() then
               ["T0_init"; "T2_S1"; "accept_S2"; "T1_S2"; "accept_S3";
                "T0_S3"; "accept_S4"
               ]
             else
               ["aorai_CurStates"]
           in
           let check_precision state =
             List.iter
               (fun auto_state ->
                 let lv =
                   Var (Globals.Vars.find_from_astinfo auto_state VGlobal),
                   NoOffset
                 in
                 let _, status =
                   !Db.Value.eval_lval ~with_alarms None state lv 
                 in
                 let _, status = Cvalue.V.find_lonely_key status in
                (* project_int will raise an exception if the value gets
                   imprecise. *)
                 ignore (Ival.project_int status)
               )
               states
           in
          (* We check the precision at the beginning of the function and
             at its end, to have more clues on where the loss of precision
             occured. *)
           (try
              List.iter check_precision state_init
            with Ival.Not_Singleton_Int ->
              M.fatal "Analysis is not precise enough to conclude. \
                      Aborting at start of function %s; backtrace:@\n%a"
                (Kernel_function.get_name kf)
                (Pretty_utils.pp_list
                   ~sep:"<-@ "
                   (fun fmt (kf,_) -> Format.pp_print_string fmt
                     (Kernel_function.get_name kf)))
                call_stack);
           (try
              List.iter check_precision state_init
            with Ival.Not_Singleton_Int ->
              M.warning "Analysis is not precise enough to conclude. \
                    At end of function %s; backtrace:@\n%a"
                (Kernel_function.get_name kf)
                (Pretty_utils.pp_list
                   ~sep:"<-@ "
                   (fun fmt (kf,_) -> Format.pp_print_string fmt
                     (Kernel_function.get_name kf)))
                call_stack)
         with Not_found -> M.fatal "Got not found exception"
       with Exit -> ())
       end

(* library functions that need renaming when using Cute or PathCrawler *)
let rename = [ "memcpy"; "memset"; "strcpy"; "strlen" ]

(* Clean up the AST from Frama-C specific built-ins
   before passing the code to Cute or PathCrawler *)
class elim_frama_c prj =
object(self)
  inherit Visitor.frama_c_copy prj

  method vglob_aux = function
    | GFun(f,_) ->
	let to_find = [ "Frama_C_nondet";
                        "Frama_C_interval" ; "__Frama_C_malloc_at_pos" ]
	in let is_recognized () = List.mem f.svar.vname to_find
	in if is_recognized () then ChangeTo []
	else DoChildren
    | GVarDecl (_,vinfo,_) ->
	let to_find =
          [ "Frama_C_nondet";
            "Frama_C_interval" ; "__Frama_C_malloc_at_pos" ;
	    "Frama_C_update_entropy" ; "Frama_C_alloc_infinite" ]
	in let is_recognized () = List.mem vinfo.vname to_find
	in if is_recognized () then ChangeTo []
	else DoChildren
    | GVar (vinfo,_,_) ->
	let to_find = [ "Frama_C_entropy_source" ]
	in let is_recognized () = List.mem vinfo.vname to_find
	in if is_recognized () then ChangeTo []
	else DoChildren
    | _ -> SkipChildren
end

(* Remove dead code, that is code that the value analysis has not visited.
   This visitor also renames libc functions which would also conflicts with
   Cute or PathCrawler instrumentation.
 *)
class dead_code_elim prj =
  let behavior = Cil.copy_visit prj in
object(self)
  inherit Visitor.generic_frama_c_visitor behavior

  method vfile f =
    ChangeDoChildrenPost(f, fun f -> Cil.reset_behavior_varinfo behavior; f)

  method vglob_aux = function
      GFun(f,_) ->
        (* if the function is not called, and its address not taken, we
           do not keep it.
         *)
        if
          !Db.Value.is_called (Globals.Functions.get f.svar) ||
            f.svar.vaddrof
        then DoChildren
        else ChangeTo []
    | GType _ | GCompTag _ | GCompTagDecl _
    | GEnumTag _ | GEnumTagDecl _ | GVarDecl _
    | GVar _ | GAsm _ | GPragma _ | GAnnot _ | GText _
        -> JustCopy

  method vfunc f =
    let is_name_changed () =
      let ret =
	List.mem f.svar.vname rename
      in
      if ret then
	M.feedback
          ~level:2 "Function %s => name changed\n" f.svar.vname;
      ret
    in
    let change_name n =
      (* avoid clashes with Cute or PathCrawler *)
      M.feedback ~level:2 "Function %s => %s\n" n.svar.vname
        ("Calmos_" ^ n.svar.vname);
      n.svar.vname <- "Calmos_" ^ n.svar.vname ;
      n
    in
    let v = Cil.get_original_varinfo self#behavior f.svar in
    if !Db.Value.is_called (Globals.Functions.get v) then
      if is_name_changed () then
	ChangeDoChildrenPost(f, change_name)
      else
	DoChildren
    else if f.svar.vaddrof then
      (* The function is not called, but its address is taken in some
         structure. We need to keep it, but with an empty body (Cute and
         PathCrawler would complain if we only keep an external declaration)
       *)
      let typ = Cil.getReturnType f.svar.vtype in
      M.feedback ~level:2 "Function %a, return type: %a"
        Printer.pp_varinfo f.svar Printer.pp_typ typ;
      let valid_sid = true in
      let loc = Cil_const.CurrentLoc.get () in
      let stmt =
        if Cil.isVoidType typ then
          Cil.mkStmt ~valid_sid (Return (None, Cil_const.CurrentLoc.get()))
        else (
	  M.feedback ~level:2
            "Function %a, return type: %a considered as NON void type"
	    Printer.pp_varinfo f.svar Printer.pp_typ typ;

          Cil.mkStmt ~valid_sid
            (Return (Some (Cil.zero ~loc), loc))
	)
      in
      f.sbody <- Cil.mkBlock [stmt] ;
      if is_name_changed () then JustCopyPost change_name
      else JustCopy
    else assert false (* should not happen *)

  method vstmt_aux stmt = match stmt.skind with
      (* Remove instructions that are not reached. *)
      Instr _ | Goto _ | Break _ | Continue _ ->
        if Db.Value.is_accessible
          (Kstmt (Cil.get_original_stmt self#behavior stmt))
        then DoChildren
        else ChangeTo (mkEmptyStmt ~valid_sid:true ())
    | If _ ->
        (* If both branches are dead, remove the test itself. *)
        let change_if s =
          match s.skind with
              If (_,{bstmts = []} ,{bstmts = []},_) ->
                s.skind <- Instr (Skip (Cil_datatype.Stmt.loc s)); s
            | _ -> s
        in ChangeDoChildrenPost(stmt,change_if)
    | Block b ->
        let change_block s =
          match s.skind with
            | Block { bstmts = []} ->
                (* empty block. Replace by Skip and keep labels if
                   any. *)
              s.skind <- Instr (Skip (Cil_datatype.Stmt.loc s)); s
            | _ -> s
        in ChangeDoChildrenPost(stmt,change_block)
    | Return _ | Loop _ | Switch _ | UnspecifiedSequence _
    | TryFinally _ | TryExcept _ -> DoChildren
        (* Could be a bit more clever when filtering entire dead
           blocks, but this should be sufficient for now...  *)
  method vblock b =
    (* After processing block, remove Skip unless they have a label
       attached to them. *)
    let compress_block b =
      let l = List.fold_left
        (fun aux s ->
           match s.skind,s.labels, Annotations.code_annot s with
               Instr (Skip _),[],[] -> aux
             | _ -> s::aux)
        [] b.bstmts
      in b.bstmts <- List.rev l;
      b
    in
    ChangeDoChildrenPost(b,compress_block)
end

(* Change the return value of a function. An invariant of the AST is
   that all function have an unique return statement of the form
   return __retres.  This is used to change the value of the maximum
   number of steps allowed for the main loop and to enable update of
   hardware status. The initial value analysis does only one step of
   the loop, without changing hardware status, as it only aims at
   removing function pointers. The real analysis must perform several
   steps, with a varying status.  *)
class change_retres expr =
object(self)
  inherit Visitor.frama_c_inplace
  method vinst =
    function
        Set ((Var {vname = "__retres"},NoOffset) as lv,_,loc) ->
          ChangeTo([Set(lv,expr,loc)])
      | _ -> SkipChildren
end

let change_retres f expr =
  let f = Kernel_function.get_definition f in
  ignore (Visitor.visitFramacBlock (new change_retres expr) f.sbody)

(* inline a function call. This way, the states corresponding to various
   execution paths won't be merged when the function returns, but instead will
   be kept separated in the caller. Allows for a more quick analysis in the
   initial phase, will retaining some precision in the final analysis.
*)
class copy =
  object(self)
    inherit Visitor.frama_c_copy (Project.current())
      method vstmt_aux s =
        (* refresh the id to avoid clash when multiple inlinings
           occur. *)
        s.sid <- Cil.Sid.next ();
        match s.skind with
            (* there's no return anymore. *)
            Return _ ->
              s.skind <- Instr (Skip (Cil_const.CurrentLoc.get()));
              SkipChildren
          | _ -> DoChildren
      method vvdec v =
        (* refresh the id *)
        Cil_const.set_vid v; DoChildren
  end

class inline_call f body =
  object(self)
    inherit Visitor.frama_c_inplace
    method vstmt_aux s = match s with
    { skind = Instr(Call(_,{enode = Lval(Var v,NoOffset)},_,_)) }
      when v.vname = f ->
        (* works on a fresh copy of the body. *)
        let body = Visitor.visitFramacFunction (new copy) body in
        let curr = Extlib.the self#current_func in
        (* local variables of the inlined function must be put
           in the caller.
         *)
        curr.slocals <- curr.slocals @ body.slocals;
        ChangeTo (Cil.mkStmt ~valid_sid:true (Block body.sbody))
      | _ -> DoChildren
  end

let inline_call f body func =
  ignore (Visitor.visitFramacFunction (new inline_call f body) func)

(* launch WP to check that the automaton is deterministic *)
let check_deterministic_lemmas () =
  let compute f _ annot =
    match annot with
      | Dlemma(name,false,_,_,_,_) ->
        let ip =
          Extlib.the (Property.ip_of_global_annotation_single annot)
        in
        f name ip
      | _ -> ()
  in
  let wp_compute =
    Dynamic.get ~plugin:"Wp" "wp_compute_ip"
      (Datatype.func Property.ty Datatype.unit)
  in
  let prove _ ip = wp_compute ip in
  let check s ip =
    match Property_status.Feedback.get ip with
      | Never_tried -> M.warning "WP did not try to prove %s" s
      | Considered_valid | Valid | Valid_under_hyp -> ()
      | Unknown -> M.warning "Lemma %s could not be proved" s
      | Invalid | Invalid_under_hyp ->
        M.fatal "Lemma %s is invalid. Automaton is not deterministic" s
      | Invalid_but_dead | Valid_but_dead | Unknown_but_dead ->
        M.warning "Lemma %s marked as dead" s
      | Inconsistent -> M.warning "Lemma %s has inconsistent status" s
  in
  Annotations.iter_global (compute prove);
  Annotations.iter_global (compute check)

(***** The main body of the analysis *****)

let run () =
  let default = Project.current() in
  let tstart = Unix.times () in
  let tinit_parse = ref tstart in
  let tinit_value = ref tstart in
  let tdead_code = ref tstart in
  let tconstant_propagation = ref tstart in
  let tbefore_aorai = ref tstart in
  let tparse_aorai = ref tstart in
  if not (OnlyFinalAnalysis.get()) then begin
    let name = "leon_propagated_constants.i" in
    (* If sources have not been modified, we start directly with the
       scenario, avoiding a costly first step of value analysis, except
       if -force is given.  *)
    let modified_src =
      List.fold_left
        (fun max x ->
           let cur = (Unix.stat x).Unix.st_mtime in
           try if max < cur then cur else max
           with Unix.Unix_error (e,_,_) ->
             M.fatal "Unable to stat source file %s:@\n%s" x
               (Unix.error_message e))
        0.0 ("main.c"::"script.ml" :: src_files ())
    in
    let modified_target=
      try
        (Unix.stat name).Unix.st_mtime
      with Unix.Unix_error _ -> modified_src
    in
    if Force.get () || modified_target <= modified_src then begin
      config ();
      M.feedback "starting initial computation";
      Ast.compute();
      tinit_parse := Unix.times ();
      !Db.Value.compute ();
      tinit_value := Unix.times ();
      M.feedback "initial computation done";
       (* we periodically save the state of the analyzer, so that it
          can be inspected by frama-c-gui in case something went
          wrong. *)
       Project.save_all "driverInit.sav";
      (* elimination of the parts of the code that have not been reached
         by the initial analysis. *)
      let vis prj = new dead_code_elim prj in
      M.result "dead code elimination started";
      let dead_code_elim =
        File.create_project_from_visitor "dead code elim" vis
      in
      (* Constant propagation. *)
      tdead_code:=Unix.times ();
      Project.set_current dead_code_elim;
      value_config();
      Project.save_all "driverDCE.sav" ;
      let prj =
        !Db.Constant_Propagation.get
          Datatype.String.Set.empty ~cast_intro:true
      in
      keep_calmos_options default prj;
      tconstant_propagation:=Unix.times();
      Project.save_all "driverCP.sav" ;
      Project.set_current prj ;
      (* Changes the parameters to keep separated states for various
         hardware status and augment the number of steps for the main loop.
       *)
      let max_step = Globals.Functions.find_by_name "max_step" in
      let uart_random_status =
        Kernel_function.get_definition
          (Globals.Functions.find_by_name "uart_random_status")
      in
      let main =
        Kernel_function.get_definition
          (Globals.Functions.find_by_name "main")
      in
      let loc = (Kernel_function.get_vi max_step).vdecl in
      change_retres max_step (Cil.integer ~loc 5);
      inline_call "uart_random_status" uart_random_status main;
      M.feedback "ok";
      (* If we use Cute or PathCrawler, clean up references to
         Frama-C internal built-ins. Otherwise, proceed directly
         to ltl-to-acsl phase.  *)
      if not (FramaCPathFinder.get ()) then begin
        let fvis prj = new elim_frama_c prj
        in let final_elim =
	  File.create_project_from_visitor "final_elim" fvis
        in
        Project.remove ~project:prj ();
        Project.set_current final_elim
      end;
      let chan = open_out name in
      let fmt = Format.formatter_of_out_channel chan in
      File.pretty_ast ~fmt ();
      close_out chan;
      Project.save_all "driverLTL.sav";
      (* remove the current project, whatever it is. *)
      Project.remove ()
    end else begin
      M.result
        "No need to re-create %s. Starting directly aorai"
        name;
    end;
    tbefore_aorai := Unix.times();
    (* Main analysis: instrumentation of the code through aorai. *)
    let ltl_prj = Project.create "aorai instrumentation" in
    keep_calmos_options default ltl_prj;
    Project.set_current ltl_prj;
    (* set input files. *)
    Kernel.CppCommand.set
      ("gcc -CC -E -I" ^ libc_path ^" -I." )
    ;

    let files =
      if FramaCPathFinder.get () then [name ]
      else [name ; "cute_hook.h" ]
    in
    (* If a dot representation of the automaton already exists,
       remove it. Otherwise aorai keeps generating new files
       instead of overwriting.
     *)
    (try Unix.unlink ((Filename.chop_suffix name ".i") ^ ".dot")
     with Unix.Unix_error _ -> ());
    Kernel.Files.set files;
    Ast.compute ();
    tparse_aorai:=Unix.times();
    (* set parameters for aorai *)
    if UseLtl.get() then
      Dynamic.Parameter.String.set "-aorai-ltl" "scenario.ltl"
    else
      Dynamic.Parameter.String.set "-aorai-automata" "scenario.ya";
    Dynamic.Parameter.String.set "-aorai-output-c-file" "scenario.c";
    Dynamic.Parameter.Bool.set "-aorai-dot" true;
    Dynamic.Parameter.Bool.set "-aorai-AI-off" true;
    Dynamic.get ~plugin:"Aorai" "run"
      (Datatype.func Datatype.unit Datatype.unit) () ;
    (* If we use Cute or PathCrawler, this is it. Otherwise, launch the
       main value analysis.
     *)
  end;
  if FramaCPathFinder.get () then begin
    let tafter_aorai = Unix.times () in
    let find_path_project = Project.create "try to find a path" in
    keep_calmos_options default find_path_project;
    Project.set_current find_path_project;
    Kernel.Files.set ["scenario.c"; "calmos_hook.c" ];
    Kernel.PreprocessAnnot.set true;
    Ast.compute();
    let tparse = Unix.times () in
    (* If we are using the deterministic automaton, use wp to check the
       lemmas generated by Aorai *)
    if not (UseLtl.get()) then
      Project.on find_path_project check_deterministic_lemmas ();
    let tprove = Unix.times () in
    (* inspects value of alarm state at the end of each step. *)
    Db.Value.Record_Value_Superposition_Callbacks.extend
      (inspect_automaton_state());
    value_config();
    if Use_builtins.get () then begin
      Dynamic.Parameter.StringSet.add
        "-val-builtin" "malloc:Frama_C_alloc_by_stack";
      Dynamic.Parameter.StringSet.add 
        "-val-builtin" "memcpy:Frama_C_memcpy";
      Dynamic.Parameter.StringSet.add "-val-builtin" "memset:Frama_C_memset";
    end;
    (* launch value analysis *)
    try
      !Db.Value.compute ();
    with Cmdline.Exit ->
      let tfinal = Unix.times () in
      let print_time s t1 t2 =
        M.feedback ~dkey:dkey_time "time taken for %s: user %.3fs, sys %.3fs"
          s
          (t2.Unix.tms_utime -. t1.Unix.tms_utime) 
          (t2.Unix.tms_stime -. t1.Unix.tms_stime)
      in
      print_time "initial parsing" tstart !tinit_parse;
      print_time "initial value analysis" !tinit_parse !tinit_value;
      print_time "dead code elimination" !tinit_value !tdead_code;
      print_time "constant propagation" !tdead_code !tconstant_propagation;
      print_time "parsing of constant propagation" !tbefore_aorai !tparse_aorai;
      print_time "aorai transformation" !tparse_aorai tafter_aorai;
      print_time "parsing" tafter_aorai tparse;
      print_time "proving" tparse tprove;
      print_time "value  " tprove tfinal
  end;
  (* we're done. Tell Frama-C not to perform other
     operations after that. *)
  raise Cmdline.Exit
;;

(* register us in the main function of Frama-C. *)
Db.Main.extend run;;
