#include "stdlib.h"
#include "stubs.h"
#include "termbits.h"

/* PH (AJOUT1) */
/* PAGE_SIZE defined in stubs.h */
void *get_zeroed_page(unsigned flags) { return kzalloc((size_t) PAGE_SIZE, flags); }
/* fin (AJOUT1) */

void panic(const char* fmt,...) { while(1); }

void init_waitqueue_head(wait_queue_head_t *q)
{
        spin_lock_init(&q->lock);
        INIT_LIST_HEAD(&q->task_list);
}

static int kobject_set_name_vargs(struct kobject *kobj, const char *fmt)
{
  char *name = kmalloc(strlen(fmt)+1,0);

  if (!name)
    return -ENOMEM;

  strcpy(name,fmt);

  /* Free the old name, if necessary. */
  if (kobj->name) kfree(kobj->name);

  /* Now, set the new name */
  kobj->name = name;

  return 0;
}

#define kobject_set_name kobject_set_name_vargs

static void kobj_kset_join(struct kobject *kobj)
{
        if (!kobj->kset)
                return;

        kset_get(kobj->kset);
        spin_lock(&kobj->kset->list_lock);
        list_add_tail(&kobj->entry, &kobj->kset->list);
        spin_unlock(&kobj->kset->list_lock);
}

static void kobj_kset_leave(struct kobject *kobj)
{
        if (!kobj->kset)
                return;

        spin_lock(&kobj->kset->list_lock);
        list_del_init(&kobj->entry);
        spin_unlock(&kobj->kset->list_lock);
        kset_put(kobj->kset);
}


static int kobject_add_internal(struct kobject *kobj)
{
        int error = 0;
        struct kobject *parent;

        if (!kobj)
                return -ENOENT;

        if (!kobj->name || !kobj->name[0]) {
                pr_debug("kobject: (%p): attempted to be registered with empty "
                         "name!\n", kobj);
                WARN_ON(1);
                return -EINVAL;
        }

        parent = kobject_get(kobj->parent);

        /* join kset if set, use it as parent if we do not already have one */
        if (kobj->kset) {
                if (!parent)
                        parent = kobject_get(&kobj->kset->kobj);
                kobj_kset_join(kobj);
                kobj->parent = parent;
        }

        pr_debug("kobject: '%s' (%p): %s: parent: '%s', set: '%s'\n",
                 kobject_name(kobj), kobj, __FUNCTION__,
                 parent ? kobject_name(parent) : "<NULL>",
                 kobj->kset ? kobject_name(&kobj->kset->kobj) : "<NULL>");

        error = create_dir(kobj);
        if (error) {
                kobj_kset_leave(kobj);
                kobject_put(parent);
                kobj->parent = NULL;

                /* be noisy on error issues */
                if (error == -EEXIST)
                        printk(KERN_ERR "%s failed for %s with "
                               "-EEXIST, don't try to register things with "
                               "the same name in the same directory.\n",
                               __FUNCTION__, kobject_name(kobj));
                else
                        printk(KERN_ERR "%s failed for %s (%d)\n",
                               __FUNCTION__, kobject_name(kobj), error);
                dump_stack();
        } else
                kobj->state_in_sysfs = 1;

        return error;
}

static int kobject_add_varg(struct kobject *kobj, struct kobject *parent,
                            const char* fmt)
{
  int retval ;

        retval = kobject_set_name_vargs(kobj,fmt);
        if (retval) {
                printk(KERN_ERR "kobject: can not set name properly!\n");
                return retval;
        }
        kobj->parent = parent;
        return kobject_add_internal(kobj);
}


int kobject_add(struct kobject *kobj, struct kobject *parent,const char* fmt)
{
        int retval;

        if (!kobj)
                return -EINVAL;

        if (!kobj->state_initialized) {
                printk(KERN_ERR "kobject '%s' (%p): tried to add an "
                       "uninitialized object, something is seriously wrong.\n",
                       kobject_name(kobj), kobj);
                //dump_stack();
                return -EINVAL;
        }
        retval = kobject_add_varg(kobj, parent,fmt);

        return retval;
}


void kref_init(struct kref* kref) { kref->refcount.counter = 1; }

void kref_get(struct kref* kref) { kref->refcount.counter++; }

int kref_put(struct kref *kref, void (*release)(struct kref *kref))
  {
    kref->refcount.counter--;
    //VP triggers recursive call
    //if (!kref->refcount.counter) { release(kref); return 1; }
    return 0;
  }

void klist_init(struct klist * k, void (*get)(struct klist_node *),
                void (*put)(struct klist_node *))
{
        INIT_LIST_HEAD(&k->k_list);
        spin_lock_init(&k->k_lock);
        k->get = get;
        k->put = put;
}

static void add_head(struct klist * k, struct klist_node * n)
{
        spin_lock(&k->k_lock);
        list_add(&n->n_node, &k->k_list);
        spin_unlock(&k->k_lock);
}

static void add_tail(struct klist * k, struct klist_node * n)
{
        spin_lock(&k->k_lock);
        list_add_tail(&n->n_node, &k->k_list);
        spin_unlock(&k->k_lock);
}


static void klist_node_init(struct klist * k, struct klist_node * n)
{
        INIT_LIST_HEAD(&n->n_node);
        init_completion(&n->n_removed);
        kref_init(&n->n_ref);
        n->n_klist = k;
        if (k->get)
                k->get(n);
}


/**
 *      klist_add_head - Initialize a klist_node and add it to front.
 *      @n:     node we're adding.
 *      @k:     klist it's going on.
 */

void klist_add_head(struct klist_node * n, struct klist * k)
{
        klist_node_init(k, n);
        add_head(k, n);
}

EXPORT_SYMBOL_GPL(klist_add_head);


/**
 *      klist_add_tail - Initialize a klist_node and add it to back.
 *      @n:     node we're adding.
 *      @k:     klist it's going on.
 */

void klist_add_tail(struct klist_node * n, struct klist * k)
{
        klist_node_init(k, n);
        add_tail(k, n);
}

EXPORT_SYMBOL_GPL(klist_add_tail);


//static void klist_release(struct kref * kref)
//{
//        struct klist_node * n = container_of(kref, struct klist_node, n_ref);
//
//        list_del(&n->n_node);
//        complete(&n->n_removed);
//        n->n_klist = NULL;
//}

//static int klist_dec_and_del(struct klist_node * n)
//{
//       return kref_put(&n->n_ref, klist_release);
//}


/**
 *      klist_del - Decrement the reference count of node and try to remove.
 *      @n:     node we're deleting.
 */

//void klist_del(struct klist_node * n)
//{
//        struct klist * k = n->n_klist;
//        void (*put)(struct klist_node *) = k->put;

//        spin_lock(&k->k_lock);
//       if (!klist_dec_and_del(n))
//                put = NULL;
//        spin_unlock(&k->k_lock);
//        if (put)
//                put(n);
//}

//void klist_remove(struct klist_node * n)
//{
//        klist_del(n);
//        wait_for_completion(&n->n_removed);
//}

/**
 *      klist_node_attached - Say whether a node is bound to a list or not.
 *      @n:     Node that we're testing.
 */

int klist_node_attached(struct klist_node * n)
{
        return (n->n_klist != NULL);
}


initcall_t __con_initcall_start[1];
initcall_t* __con_initcall_end = __con_initcall_start;

struct kobj_map {
	struct probe {
		struct probe *next;
		dev_t dev;
		unsigned long range;
		struct module *owner;
		kobj_probe_t *get;
		int (*lock)(dev_t, void *);
		void *data;
	} *probes[255];
	struct mutex *lock;
};

static struct kobject *cdev_get(struct cdev *p)
{
  //NdV: this just increments some counters in complicated data structures.
  //	struct module *owner = p->owner;
  //	struct kobject *kobj;

  //	if (owner && !try_module_get(owner))
  //		return NULL;
  //	kobj = kobject_get(&p->kobj);
  //	if (!kobj)
  //		module_put(owner);
  //return kobj;
  return &p->kobj;
}

struct kobject *kobject_get(struct kobject *kobj)
{
        if (kobj)
                kref_get(&kobj->kref);
        return kobj;
}


static inline struct kobj_type *get_ktype(struct kobject *kobj)
{
        return kobj->ktype;
}
static inline int kobject_uevent(struct kobject *kobj,
                                 enum kobject_action action)
{ return 0; }

void kobject_del(struct kobject *kobj)
{
        if (!kobj)
                return;

        //        sysfs_remove_dir(kobj);
        kobj->state_in_sysfs = 0;
        kobj_kset_leave(kobj);
        kobject_put(kobj->parent);
        kobj->parent = NULL;
}

static void kobject_cleanup(struct kobject *kobj)
{
        struct kobj_type *t = get_ktype(kobj);
        const char *name = kobj->name;

        pr_debug("kobject: '%s' (%p): %s\n",
                 kobject_name(kobj), kobj, __FUNCTION__);

        if (t && !t->release)
                pr_debug("kobject: '%s' (%p): does not have a release() "
                         "function, it is broken and must be fixed.\n",
                         kobject_name(kobj), kobj);

        /* send "remove" if the caller did not do it but sent "add" */
        if (kobj->state_add_uevent_sent && !kobj->state_remove_uevent_sent) {
                pr_debug("kobject: '%s' (%p): auto cleanup 'remove' event\n",
                         kobject_name(kobj), kobj);
                kobject_uevent(kobj, KOBJ_REMOVE);
        }

        /* remove from sysfs if the caller did not do it */
        if (kobj->state_in_sysfs) {
                pr_debug("kobject: '%s' (%p): auto cleanup kobject_del\n",
                         kobject_name(kobj), kobj);
                kobject_del(kobj);
        }

	/* PH : (MODIF2) code retiré */

         if (t && t->release) { 
                 pr_debug("kobject: '%s' (%p): calling ktype release\n", 
                          kobject_name(kobj), kobj); 
                 t->release(kobj); 
         } 

        /* PH : fin (MODIF2) */

        /* free name if we allocated it */
        if (name) {
                pr_debug("kobject: '%s': free name\n", name);
                kfree(name);
        }
}

static void kobject_release(struct kref *kref)
{
        kobject_cleanup(container_of(kref, struct kobject, kref));
}


void kobject_put(struct kobject *kobj)
{
  // Same as kobject_get
  	if (kobj)
  		kref_put(&kobj->kref, kobject_release);
}

static void kobject_init_internal(struct kobject *kobj)
{
        if (!kobj)
                return;
        //kref_init(&kobj->kref);
        INIT_LIST_HEAD(&kobj->entry);
        kobj->state_in_sysfs = 0;
        kobj->state_add_uevent_sent = 0;
        kobj->state_remove_uevent_sent = 0;
        kobj->state_initialized = 1;
}


void kobject_init(struct kobject *kobj, struct kobj_type *ktype)
{
        char *err_str;

        if (!kobj) {
                err_str = "invalid kobject pointer!";
                goto error;
        }
        if (!ktype) {
                err_str = "must have a ktype to be initialized properly!\n";
                goto error;
        }
        if (kobj->state_initialized) {
                /* do not error out as sometimes we can recover */
                printk(KERN_ERR "kobject (%p): tried to init an initialized "
                       "object, something is seriously wrong.\n", kobj);
                dump_stack();
        }

        kobject_init_internal(kobj);
        kobj->ktype = ktype;
        return;

error:
        printk(KERN_ERR "kobject (%p): %s\n", kobj, err_str);
        dump_stack();
}

struct device *get_device(struct device *dev)
{
        return dev ? to_dev(kobject_get(&dev->kobj)) : NULL;
}

static struct kobject *get_device_parent(struct device *dev,
                                         struct device *parent)
{
        /* class devices without a parent live in /sys/class/<classname>/ */
        if (dev->class && (!parent || parent->class != dev->class))
                return &dev->class->subsys.kobj;
        /* all other devices keep their parent */
        else if (parent)
                return &parent->kobj;

        return NULL;
}

static void setup_parent(struct device *dev, struct device *parent)
{
        struct kobject *kobj;
        kobj = get_device_parent(dev, parent);
        if (kobj)
                dev->kobj.parent = kobj;
}


void cdev_put(struct cdev *p)
{
	if (p) {
		struct module *owner = p->owner;
		kobject_put(&p->kobj);
		module_put(owner);
	}
}

static struct kobj_map *cdev_map;

int kobj_map(struct kobj_map *domain, dev_t dev, unsigned long range,
             struct module *module, kobj_probe_t *probe,
             int (*lock)(dev_t, void *), void *data)
{
        unsigned n = MAJOR(dev + range - 1) - MAJOR(dev) + 1;
        unsigned index = MAJOR(dev);
        unsigned i;
        struct probe *p;

        if (n > 255)
                n = 255;

        p = kmalloc(sizeof(struct probe) * n, GFP_KERNEL);

        if (p == NULL)
                return -ENOMEM;

        for (i = 0; i < n; i++, p++) {
                p->owner = module;
                p->get = probe;
                p->lock = lock;
                p->dev = dev;
                p->range = range;
                p->data = data;
        }
        // mutex_lock(domain->lock);
        for (i = 0, p -= n; i < n; i++, p++, index++) {
                struct probe **s = &domain->probes[index % 255];
                while (*s && (*s)->range < range)
                        s = &(*s)->next;
                p->next = *s;
                *s = p;
        }
        //mutex_unlock(domain->lock);
        return 0;
}

static struct kobject *exact_match(dev_t dev, int *part, void *data)
{
	struct cdev *p = data;
	return &p->kobj;
}

static int exact_lock(dev_t dev, void *data)
{
	struct cdev *p = data;
	return cdev_get(p) ? 0 : -1;
}

int cdev_add(struct cdev *p, dev_t dev, unsigned count)
 {
        p->dev = dev;
        p->count = count;
        return kobj_map(cdev_map, dev, count, NULL, exact_match, exact_lock, p);
        return 0;
 }

int register_chrdev_region(dev_t d, unsigned u, const char *c) { return 0; }

loff_t no_llseek(struct file *file, loff_t offset, int origin)
{ return -ESPIPE; }

static DEFINE_SPINLOCK(cdev_lock);

static void cdev_purge(struct cdev *cdev)
{
	spin_lock(&cdev_lock);
	while (!list_empty(&cdev->list)) {
		struct inode *inode;
		inode = container_of(cdev->list.next, struct inode, i_devices);
		list_del_init(&inode->i_devices);
		inode->i_cdev = NULL;
	}
	spin_unlock(&cdev_lock);
}

static void cdev_default_release(struct kobject *kobj)
{
	struct cdev *p = container_of(kobj, struct cdev, kobj);
	cdev_purge(p);
}

static struct kobj_type ktype_cdev_default = {
	.release	= cdev_default_release,
};

void cdev_init(struct cdev *cdev, const struct file_operations *fops)
{
        memset(cdev, 0, sizeof *cdev);
        INIT_LIST_HEAD(&cdev->list);
        kobject_init(&cdev->kobj, &ktype_cdev_default);
        cdev->ops = fops;
}

static struct kobject *base_probe(dev_t dev, int *part, void *data)
{
  //	if (request_module("char-major-%d-%d", MAJOR(dev), MINOR(dev)) > 0)
		/* Make old-style 2.4 aliases work */
  //		request_module("char-major-%d", MAJOR(dev));
	return NULL;
}

struct kobj_map *kobj_map_init(kobj_probe_t *base_probe, struct mutex *lock)
{
	struct kobj_map *p = kmalloc(sizeof(struct kobj_map), GFP_KERNEL);
	struct probe *base = kzalloc(sizeof(*base), GFP_KERNEL);
	int i;

	if ((p == NULL) || (base == NULL)) {
		kfree(p);
		kfree(base);
		return NULL;
	}

	base->dev = 1;
	base->range = ~0;
	base->get = base_probe;
	for (i = 0; i < 255; i++)
		p->probes[i] = base;
	p->lock = lock;
	return p;
}

static DEFINE_MUTEX(chrdevs_lock);

void __init chrdev_init(void)
{
	cdev_map = kobj_map_init(base_probe, &chrdevs_lock);
	//bdi_init(&directly_mappable_cdev_bdi);
}

struct kobject *kobj_lookup(struct kobj_map *domain, dev_t dev, int *index)
{
	struct kobject *kobj;
	struct probe *p;
	unsigned long best = ~0UL;

retry:
	//mutex_lock(domain->lock);
	for (p = domain->probes[MAJOR(dev) % 255]; p; p = p->next) {
		struct kobject *(*probe)(dev_t, int *, void *);
		struct module *owner;
		void *data;

		if (p->dev > dev || p->dev + p->range - 1 < dev)
			continue;
		if (p->range - 1 >= best)
			break;
		if (!try_module_get(p->owner))
			continue;
		owner = p->owner;
		data = p->data;
		probe = p->get;
		best = p->range - 1;
		*index = dev - p->dev;
		if (p->lock && p->lock(dev, data) < 0) {
			module_put(owner);
			continue;
		}
		//mutex_unlock(domain->lock);
		kobj = probe(dev, index, data);
		/* Currently ->owner protects _only_ ->probe() itself. */
		module_put(owner);
		if (kobj)
			return kobj;
		goto retry;
	}
	//mutex_unlock(domain->lock);
	return NULL;
}

static int chrdev_open(struct inode *inode, struct file *filp)
{
	struct cdev *p;
	struct cdev *new = NULL;
	int ret = 0;

	spin_lock(&cdev_lock);
	p = inode->i_cdev;
	if (!p) {
		struct kobject *kobj;
		int idx;
		spin_unlock(&cdev_lock);
		kobj = kobj_lookup(cdev_map, inode->i_rdev, &idx);
		if (!kobj)
			return -ENXIO;
		new = container_of(kobj, struct cdev, kobj);
		spin_lock(&cdev_lock);
		p = inode->i_cdev;
		if (!p) {
			inode->i_cdev = p = new;
			inode->i_cindex = idx;
			list_add(&inode->i_devices, &p->list);
			new = NULL;
		} else if (!cdev_get(p))
			ret = -ENXIO;
	} else if (!cdev_get(p))
		ret = -ENXIO;
	spin_unlock(&cdev_lock);
	cdev_put(new);
	if (ret)
		return ret;
	filp->f_op = fops_get(p->ops);
	if (!filp->f_op) {
		cdev_put(p);
		return -ENXIO;
	}
	if (filp->f_op->open) {
		lock_kernel();
		ret = filp->f_op->open(inode,filp);
		unlock_kernel();
	}
	if (ret)
		cdev_put(p);
	return ret;
}

const struct file_operations def_chr_fops = {
	.open = &chrdev_open,
};

int nonseekable_open(struct inode *inode, struct file *filp)
{
        filp->f_mode &= ~(FMODE_LSEEK | FMODE_PREAD | FMODE_PWRITE);
        return 0;
}
#if 0
#define __ATTR(_name,_mode,_show,_store) { \
        .attr = {.name = __stringify(_name), .mode = _mode },   \
        .show   =  _show,                                        \
        .store  =  _store,                                       \
}
#else
#define  __ATTR(_name,_mode,_show,_store) { \
    .attr = {.name = __stringify(_name), .mode = _mode } }
#endif

static struct device_attribute uevent_attr =
        __ATTR(uevent, S_IRUGO | S_IWUSR, show_uevent, store_uevent);

static struct device_attribute devt_attr =
        __ATTR(dev, S_IRUGO, show_dev, NULL);


int device_add(struct device *dev)
{
        struct device *parent = NULL;
        struct class_interface *class_intf;
        int error;

        dev = get_device(dev);
        if (!dev || !strlen(dev->bus_id)) {
                error = -EINVAL;
                goto Done;
        }

        //pr_debug("device: '%s': %s\n", dev->bus_id, __FUNCTION__);

        parent = get_device(dev->parent);
        setup_parent(dev, parent);

        /* first, register with generic layer. */
        error = kobject_add(&dev->kobj, dev->kobj.parent,dev->bus_id);//, "%s", dev->bus_id);
        if (error)
                goto Error;

        /* notify platform of device entry */
	/* PH (MODIF3) : code retiré */
/*         if (platform_notify) */
/*                 platform_notify(dev); */
	/* fin (MODIF3) */

        /* notify clients of device entry (new way) */
        /*        if (dev->bus)
                blocking_notifier_call_chain(&dev->bus->p->bus_notifier,
                                             BUS_NOTIFY_ADD_DEVICE, dev);
        */
        //        error = device_create_file(dev, &uevent_attr);
        //if (error)
        //        goto attrError;

        //if (MAJOR(dev->devt)) {
        //        error = device_create_file(dev, &devt_attr);
        //        if (error)
        //                goto ueventattrError;
        //}

        //error = device_add_class_symlinks(dev);
        //if (error)
        //        goto SymlinkError;
        //error = device_add_attrs(dev);
        //if (error)
        //       goto AttrsError;
        //error = dpm_sysfs_add(dev);
        //if (error)
        //        goto PMError;
        //device_pm_add(dev);
        //error = bus_add_device(dev);
        //if (error)
        //        goto BusError;
        //kobject_uevent(&dev->kobj, KOBJ_ADD);
        //bus_attach_device(dev);
        if (parent)
                klist_add_tail(&dev->knode_parent, &parent->klist_children);

	/* PH : (MODIF4) code enlevé */
/*         if (dev->class) { */
/*                 down(&dev->class->sem); */
/*                 /\* tie the class to the device *\/ */
/*                 list_add_tail(&dev->node, &dev->class->devices); */

/*                 /\* notify any interfaces that the device is here *\/ */
/*                 list_for_each_entry(class_intf, &dev->class->interfaces, node) */
/*                         if (class_intf->add_dev) */
/*                                 class_intf->add_dev(dev, class_intf); */
/*                 up(&dev->class->sem); */
/*         } */
	/* PH : fin (MODIF4) */
 Done:
        put_device(dev);
        return error;
 BusError:
        //device_pm_remove(dev);
 PMError:
        //if (dev->bus)
        //        blocking_notifier_call_chain(&dev->bus->p->bus_notifier,
        //                                     BUS_NOTIFY_DEL_DEVICE, dev);
        //device_remove_attrs(dev);
 AttrsError:
        //device_remove_class_symlinks(dev);
 SymlinkError:
        //if (MAJOR(dev->devt))
        //        device_remove_file(dev, &devt_attr);
 ueventattrError:
        //device_remove_file(dev, &uevent_attr);
 attrError:
        //kobject_uevent(&dev->kobj, KOBJ_REMOVE);
        //kobject_del(&dev->kobj);
 Error:
        cleanup_device_parent(dev);
        if (parent)
                put_device(parent);
        goto Done;
}

struct kset *devices_kset;

static void klist_children_get(struct klist_node *n)
{
        struct device *dev = container_of(n, struct device, knode_parent);

        get_device(dev);
}

static void klist_children_put(struct klist_node *n)
{
        struct device *dev = container_of(n, struct device, knode_parent);

        put_device(dev);
}


void device_initialize(struct device *dev)
{
        dev->kobj.kset = devices_kset;
        kobject_init(&dev->kobj, &device_ktype);
        klist_init(&dev->klist_children, klist_children_get,
                   klist_children_put);
        //INIT_LIST_HEAD(&dev->dma_pools);
        INIT_LIST_HEAD(&dev->node);
        init_MUTEX(&dev->sem);
        spin_lock_init(&dev->devres_lock);
        INIT_LIST_HEAD(&dev->devres_head);
        device_init_wakeup(dev, 0);
        set_dev_node(dev, -1);
}


int device_register(struct device *dev)
{
        device_initialize(dev);
        return device_add(dev);
}

static void device_create_release(struct device *dev) { kfree(dev); }

struct device *device_create(struct class *class, struct device *parent,
                             dev_t devt, const char *fmt/*, ...*/)
{
  //        va_list args;
        struct device *dev = NULL;
        int retval = -ENODEV;

        if (class == NULL || IS_ERR(class))
                goto error;

        dev = kzalloc(sizeof(*dev), GFP_KERNEL);
        if (!dev) {
                retval = -ENOMEM;
                goto error;
        }

        dev->devt = devt;
        dev->class = class;
        dev->parent = parent;
        dev->release = device_create_release;

        //va_start(args, fmt);
        //vsnprintf(dev->bus_id, BUS_ID_SIZE, fmt, args);
        //va_end(args);
        strcpy(dev->bus_id,fmt);
        retval = device_register(dev);
        if (retval)
                goto error;

        return dev;

error:
        kfree(dev);
        return ERR_PTR(retval);
}

/* PH : (MODIF9) => ne rien faire */
void spin_lock(spinlock_t* lock) { /* *lock = 1; */ }
void spin_unlock(spinlock_t* lock) { /* *lock = -1; */ }
/* fin (MODIF9) */

unsigned long _spin_lock_irqsave(spinlock_t *lock) { spin_lock(lock); return 0;}
unsigned long _spin_unlock_irqrestore(spinlock_t *lock,unsigned long flags) {
  spin_unlock(lock); return 0;}

void tasklet_init(struct tasklet_struct *t,
                  void (*func)(unsigned long), unsigned long data){
        t->next = NULL;
        t->state = 0;
        t->count.counter=0;
        t->func = func;
        t->data = data;
}

void put_device(struct device *dev)
{
        /* might_sleep(); */
        if (dev)
                kobject_put(&dev->kobj);
}

int kset_register(struct kset *k)
{
        int err;

        if (!k)
                return -EINVAL;

        kset_init(k);
        err = kobject_add_internal(&k->kobj);
        if (err)
                return err;
        kobject_uevent(&k->kobj, KOBJ_ADD);
        return 0;
}

static int add_class_attrs(struct class *cls) { return 0; }
static struct class *class_get(struct class *cls)
{
        if (cls)
                return container_of(kset_get(&cls->subsys),
                                    struct class, subsys);
        return NULL;
}

static void class_put(struct class *cls)
{
        if (cls)
                kset_put(&cls->subsys);
}


int class_register(struct class *cls)
 {
         int error;

         pr_debug("device class '%s': registering\n", cls->name);

         INIT_LIST_HEAD(&cls->children);
         INIT_LIST_HEAD(&cls->devices);
         INIT_LIST_HEAD(&cls->interfaces);
         kset_init(&cls->class_dirs);
         init_MUTEX(&cls->sem);
         error = kobject_set_name(&cls->subsys.kobj, cls->name);
         if (error)
                 return error;

         cls->subsys.kobj.kset = class_kset;

         cls->subsys.kobj.ktype = &class_ktype;

         error = kset_register(&cls->subsys);
         if (!error) {
                 error = add_class_attrs(class_get(cls));
                 class_put(cls);
         }
         return error;
 }

static void class_create_release(struct class *cls)
{
        pr_debug("%s called for %s\n", __FUNCTION__, cls->name);
        kfree(cls);
}

static void class_device_create_release(struct class_device *class_dev)
{
        pr_debug("%s called for %s\n", __FUNCTION__, class_dev->class_id);
        kfree(class_dev);
}


struct class *class_create(struct module *owner, const char *name)
 {
         struct class *cls;
         int retval;

         cls = kzalloc(sizeof(*cls), GFP_KERNEL);
         if (!cls) {
                 retval = -ENOMEM;
                 goto error;
         }

         cls->name = name;
         cls->owner = owner;
         cls->class_release = class_create_release;
         cls->release = class_device_create_release;

         retval = class_register(cls);
         if (retval)
                 goto error;

         return cls;

 error:
         kfree(cls);
         return ERR_PTR(retval);
 }

void kset_init(struct kset *k)
{
        kobject_init_internal(&k->kobj);
        INIT_LIST_HEAD(&k->list);
        spin_lock_init(&k->list_lock);
}

/* PH : (MODIF10) => ne rien faire */
void spin_lock_init(spinlock_t* lock) { /* *lock = 0; */ }
/* fin (MODIF10) */

/* PH (MODIF13) : ajout */
speed_t tty_termios_baud_rate(struct ktermios *termios) { return termios->c_ispeed /* 38400 */; }
speed_t tty_termios_input_baud_rate(struct ktermios *termios) { return termios->c_ospeed /* 38400 */; }
/* fin (MODIF13) */


/* définitions "triviales" */

int kernel_locked() { return 1; } /* do NOT return 0 */
int wake_up(void* a) { return 0; }
int file_move(struct file *file, struct list_head *list) { return 0; }
int spin_lock_irq(void* p) { return 0; }
int spin_unlock_irq(void* p) { return 0; }
int put_pid(void* p) { return 0; }
int schedule_delayed_work(void*p, int i) { return 0; }
int INIT_DELAYED_WORK(void* a, void* b) { return 0; }
void proc_tty_register_driver(struct tty_driver *driver ) { return; }
int tasklet_schedule(void* p) { return 0; }
int remove_wait_queue(void* a,void* b) { return 0; }
int signal_pending(void* p) { return 0; }
int mutex_lock_interruptible(void* a) { return 0; }
int lockdep_set_class(void* a, void* b) { return 0; }
int request_irq(unsigned int irq, void* handler, void* a, void* devname, void* dev_id) { return 0; }
int n_tty_ioctl(struct tty_struct *tty , struct file *file ,
		unsigned int cmd , unsigned long arg ) { return 0; }
int irq_canonicalize(int x) { return x; }
int copy_to_user(void* dst, void* src, size_t n) {
  memcpy(dst,src,n);
  return 0;
}

int copy_from_user(void* dst, void* src, size_t n) {
  memcpy(dst,src,n);
  return 0;
}

int put_user(int x, int* dst) { *dst =  x; return 0; }

void schedule() { }
